"""Config module."""
import ipaddress
import itertools
import logging
import os
import re

from copy import deepcopy
from typing import Dict, Union

import yaml

from homer.devices import Device
from homer.exceptions import HomerError

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def ip_network_constructor(loader: yaml.constructor.BaseConstructor, node: str) -> Union[str,
                                                                                         ipaddress.IPv4Network,
                                                                                         ipaddress.IPv6Network,
                                                                                         ipaddress.IPv4Interface,
                                                                                         ipaddress.IPv6Interface]:
    """Casts a string into a ipaddress.ip_network or ip_interface object.

    Arguments:
        loader (yaml loader): YAML loaded on which to apply the constructor
        node: (str): string to be casted

    Returns:
        ipaddress: an IPv4 or v6 Network or Interface object
        str: if not possible, return the original string

    """
    value = loader.construct_scalar(node)
    try:
        return ipaddress.ip_network(value)
    except ValueError:
        try:
            return ipaddress.ip_interface(value)
        except ValueError as e:
            logger.debug('Casting to ip_network or ip_interface failed, defaulting to string (%s).', e)
            return value


def ip_address_constructor(loader: yaml.constructor.BaseConstructor, node: str) -> Union[str,
                                                                                         ipaddress.IPv4Address,
                                                                                         ipaddress.IPv6Address]:
    """Casts a string into a ipaddress.ip_address object.

    Arguments:
        loader (yaml loader): YAML loaded on which to apply the constructor
        node: (str): string to be casted

    Returns:
        ipaddress: an IPv4 or v6 address object
        str: if not possible, return the original string

    """
    value = loader.construct_scalar(node)
    try:
        return ipaddress.ip_address(value)
    except ValueError as e:
        logger.debug('Casting to ip_address failed, defaulting to string (%s).', e)
        return value


def load_yaml_config(config_file: str) -> Dict:
    """Parse a YAML config file and return it.

    Arguments:
        config_file (str): the path of the configuration file.

    Returns:
        dict: the parsed config or an empty dictionary if the file doesn't exists.

    Raises:
        HomerError: if failed to load the configuration.

    """
    network_re = re.compile(r"^(\d+\.\d+\.\d+\.\d+|(?:[\da-f]{0,4}:){2,7}[\da-f]{0,4})/\d+$")
    ip_re = re.compile(r"^(\d+\.\d+\.\d+\.\d+|(?:[\da-f]{0,4}:){2,7}[\da-f]{0,4})$")
    yaml.SafeLoader.add_constructor("!ip_network", ip_network_constructor)
    yaml.SafeLoader.add_implicit_resolver('!ip_network', network_re, None)
    yaml.SafeLoader.add_constructor("!ip_address", ip_address_constructor)
    yaml.SafeLoader.add_implicit_resolver('!ip_address', ip_re, None)

    config = {}  # type: Dict
    if not os.path.exists(config_file):
        return config

    try:
        with open(config_file, 'r') as fh:
            config = yaml.safe_load(fh)

    except Exception as e:  # pylint: disable=broad-except
        raise HomerError('Could not load config file {file}: {e}'.format(file=config_file, e=e)) from e

    if config is None:
        config = {}

    return config


class HierarchicalConfig:
    """Load configuration with hierarchical override based on role, site and device."""

    def __init__(self, base_path: str, *, private_base_path: str = ''):
        """Initialize the instance.

        Arguments:
            base_path (str): the base path from where the configuration files should be loaded. The configuration
                files that will be loaded, if existing, are:
                - ``common.yaml``: common key:value pairs
                - ``roles.yaml``: one key for each role with key:value pairs of role-specific configuration
                - ``sites.yaml``: one key for each site with key:value pairs of role-specific configuration
            private_base_path (str, optional): the base path from where the private configuration files should be
                loaded, with the same structure of the above ``base_path`` argument. If existing, private
                configuration files cannot have top level keys in common with the public configuration.

        """
        self._configs = {}  # type: Dict[str, Dict]
        paths = {'public': base_path, 'private': private_base_path}
        for path, name in itertools.product(paths.keys(), ('common', 'roles', 'sites')):
            if paths[path]:
                config = load_yaml_config(os.path.join(paths[path], 'config', '{name}.yaml'.format(name=name)))
            else:
                config = {}

            self._configs['{p}_{n}'.format(p=path, n=name)] = config

    def get(self, device: Device) -> Dict:
        """Get the generated configuration for a specific device instance with all the overrides resolved.

        Arguments:
            device (homer.devices.Device): the device instance.

        Raises:
            homer.exceptions.HomerError: if any top level key is present in both the private and public configuration.

        Returns:
            dict: the generated device-specific configuration dictionary. The override order is:
            ``common``, ``role``, ``site``, ``device``. Public and private configuration are merged together.

        """
        role = device.metadata.get('role', '')
        site = device.metadata.get('site', '')
        # Deepcopying the common configurations to protect from any side effect
        public = {
            **deepcopy(self._configs['public_common']),
            **deepcopy(self._configs['public_roles'].get(role, {})),
            **deepcopy(self._configs['public_sites'].get(site, {})),
            **device.config,
            **{'metadata': device.metadata, 'hostname': device.fqdn},  # Inject also FQDN and device metadata
        }
        private = {
            **self._configs['private_common'],
            **self._configs['private_roles'].get(role, {}),
            **self._configs['private_sites'].get(site, {}),
            **device.private,
        }
        if public.keys() & private.keys():
            raise HomerError('Configuration key(s) found in both public and private config: {keys}'.format(
                keys=public.keys() & private.keys()))

        return {**public, **deepcopy(private)}



"""Vulture whitelist to avoid false positives."""


class Whitelist:
    """Helper class that allows mocking Python objects."""

    def __getattr__(self, _):
        """Mocking magic method __getattr__."""
        pass


whitelist_tests = Whitelist()
whitelist_tests.unit.test_devices.TestDevices.setup_method
whitelist_tests.unit.test_devices.TestDevices.setup_method_fixture
whitelist_tests.unit.test_config.test_uncopiable_object.Uncopiable.__deepcopy__
whitelist_tests.fixtures.plugins.plugin.NetboxDeviceDataPlugin._get_netbox_device_plugin

whitelist_cli = Whitelist()
whitelist_cli.argument_parser.subparsers.required

whitelist_netbox = Whitelist()
whitelist_netbox.NetboxData._get_vlans
whitelist_netbox.NetboxDeviceData._get_virtual_chassis_members

# Needed for vulture < 0.27
whitelist_mock = Whitelist()
whitelist_mock.return_value
whitelist_mock.side_effect



"""Templates module."""
import logging
import os

from typing import Mapping

import jinja2

from homer.exceptions import HomerError


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class Renderer:
    """Load and render templates."""

    def __init__(self, base_path: str):
        """Initialize the instance.

        Arguments:
            base_path (str): the base path to initialize the Jinja2 environment with. All templates path must be
                relative to this base path.

        """
        self._env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(os.path.join(base_path, 'templates')),
            undefined=jinja2.StrictUndefined,
            trim_blocks=True,
            lstrip_blocks=True,
            cache_size=0)

    def render(self, template_name: str, data: Mapping) -> str:
        """Render a template with the given data.

        Arguments:
            template_name (str): the name of the template to load without the file extension.
            data (dict): the dictionary of variables to pass to Jinja2 for replacement.

        Raises:
            HomerError: on error.

        Returns:
            str: the rendered template on success.
            None: on failure.

        """
        template_file = '{name}.conf'.format(name=template_name)
        try:
            template = self._env.get_template(template_file)
            return template.render(data)
        except jinja2.exceptions.TemplateSyntaxError as e:
            raise HomerError('Syntax error on template {file}'.format(file=template_file)) from e
        except jinja2.exceptions.TemplateError as e:
            raise HomerError('Could not render template {file}'.format(file=template_file)) from e



"""Tests."""
import os


TESTS_BASE_PATH = os.path.realpath(os.path.dirname(__file__))


def get_fixture_path(*paths):
    """Return the absolute path of the given fixture.

    Arguments:
        *paths: arbitrary positional arguments used to compose the absolute path to the fixture.

    Returns:
        str: the absolute path of the selected fixture.

    """
    return os.path.join(TESTS_BASE_PATH, 'fixtures', *paths)



"""Junos module tests."""
from unittest import mock

import pytest

from jnpr.junos.exception import CommitError, ConfigLoadError, RpcTimeoutError, UnlockError
from lxml import etree
from ncclient.operations.errors import TimeoutExpiredError

from homer.exceptions import HomerAbortError, HomerError, HomerTimeoutError
from homer.transports import junos


COMMIT_MESSAGE = 'commit message'
ERROR_RESPONSE = """
    <rpc-reply xmlns="urn:ietf:params:xml:ns:netconf:base:1.0"
               xmlns:junos="http://xml.juniper.net/junos/16.1I0/junos"
               xmlns:nc="urn:ietf:params:xml:ns:netconf:base:1.0">
    <load-configuration-results>
    <rpc-error>
        <error-severity>warning</error-severity>
        <error-message>Error Message</error-message>
    </rpc-error>
    <rpc-error>
        <error-severity>warning</error-severity>
        <error-message>statement not found</error-message>
    </rpc-error>
    {insert}
    </load-configuration-results>
    </rpc-reply>
"""


@mock.patch('homer.transports.junos.ConnectedDevice', autospec=True)
def test_connected_device(mocked_device):
    """It should return a context manager around a connected JunOS device."""
    with junos.connected_device('device1.example.com', username='username') as device:
        mocked_device.assert_called_once_with('device1.example.com', username='username', ssh_config=None)
        assert hasattr(device, 'commit')
        assert not mocked_device.return_value.close.called

    mocked_device.return_value.close.assert_called_once_with()


@mock.patch('homer.transports.junos.JunOSDevice', autospec=True)
class TestConnectedDevice:
    """ConnectedDevice class tests."""

    def setup_method(self, _):
        """Initialize the instance."""
        # pylint: disable=attribute-defined-outside-init
        self.fqdn = 'device1.example.com'

    def test_init(self, mocked_junos_device):
        """It should connect to the device."""
        junos.ConnectedDevice(self.fqdn)
        mocked_junos_device.assert_called_once_with(host=self.fqdn, user='', port=22, ssh_config=None)

    def test_commit_ok(self, mocked_junos_device):
        """It should commit the new config."""
        mocked_junos_device.return_value.cu = mock.MagicMock(spec_set=junos.Config)
        mocked_junos_device.return_value.cu.diff.return_value = 'diff'
        device = junos.ConnectedDevice(self.fqdn)
        callback = mock.Mock()

        device.commit('config', COMMIT_MESSAGE, callback)

        callback.assert_called_once_with('device1.example.com', 'diff')
        mocked_junos_device.return_value.cu.commit.assert_called_once_with(confirm=2, comment=COMMIT_MESSAGE)

    @pytest.mark.parametrize('is_retry', (True, False))
    def test_commit_empty_diff(self, mocked_junos_device, is_retry):
        """It should skip the callback on empty diff and based on the is_retry parameter run commit_check or not."""
        mocked_junos_device.return_value.cu = mock.MagicMock(spec_set=junos.Config)
        mocked_junos_device.return_value.cu.diff.return_value = None
        device = junos.ConnectedDevice(self.fqdn)
        callback = mock.Mock()

        device.commit('config', COMMIT_MESSAGE, callback, is_retry=is_retry)

        callback.assert_not_called()
        mocked_junos_device.return_value.cu.commit.assert_not_called()
        if is_retry:
            mocked_junos_device.return_value.cu.commit_check.assert_called_once_with()
        else:
            mocked_junos_device.return_value.cu.commit_check.assert_not_called()

    def test_commit_abort(self, mocked_junos_device):
        """It should abort the commit and not log exception."""
        mocked_junos_device.return_value.cu = mock.MagicMock(spec_set=junos.Config)
        mocked_junos_device.return_value.cu.diff.return_value = 'diff'
        device = junos.ConnectedDevice(self.fqdn)
        callback = mock.Mock()
        callback.side_effect = HomerAbortError

        with pytest.raises(HomerAbortError):
            device.commit('config', COMMIT_MESSAGE, callback)

        callback.assert_called_once_with('device1.example.com', 'diff')
        mocked_junos_device.return_value.cu.commit.assert_not_called()

    def test_commit_timeout(self, mocked_junos_device):
        """It should catch the timeout exception separately."""
        mocked_junos_device.return_value.cu = mock.MagicMock(spec_set=junos.Config)
        mocked_junos_device.return_value.cu.diff.return_value = 'diff'
        mocked_junos_device.return_value.cu.commit.side_effect = RpcTimeoutError(
            mocked_junos_device, 'commit-configuration', 30)
        device = junos.ConnectedDevice(self.fqdn)
        callback = mock.Mock()

        with pytest.raises(HomerTimeoutError):
            device.commit('config', COMMIT_MESSAGE, callback)

        callback.assert_called_once_with('device1.example.com', 'diff')
        mocked_junos_device.return_value.cu.commit_check.assert_not_called()

    def test_commit_prepare_failed_load(self, mocked_junos_device):
        """It should raise if unable to load the configuration."""
        mocked_junos_device.return_value.cu = mock.MagicMock(spec_set=junos.Config)
        mocked_junos_device.return_value.cu.load.side_effect = ConfigLoadError(
            etree.XML(ERROR_RESPONSE.format(insert='')))
        device = junos.ConnectedDevice(self.fqdn)
        callback = mock.Mock()
        with pytest.raises(ConfigLoadError):
            device.commit('config', COMMIT_MESSAGE, callback)

        callback.assert_not_called()
        mocked_junos_device.return_value.cu.commit.assert_not_called()
        mocked_junos_device.return_value.cu.rollback.assert_not_called()

    def test_commit_prepare_failed_diff(self, mocked_junos_device):
        """It should raise HomerError if unable to get the diff for the commit."""
        mocked_junos_device.return_value.cu = mock.MagicMock(spec_set=junos.Config)
        mocked_junos_device.return_value.cu.diff.side_effect = ValueError
        device = junos.ConnectedDevice(self.fqdn)
        callback = mock.Mock()
        with pytest.raises(ValueError):
            device.commit('config', COMMIT_MESSAGE, callback)

        callback.assert_not_called()
        mocked_junos_device.return_value.cu.commit.assert_not_called()

    def test_commit_callback_failed(self, mocked_junos_device):
        """It should raise HomerError if the call to the callback fails."""
        mocked_junos_device.return_value.cu = mock.MagicMock(spec_set=junos.Config)
        mocked_junos_device.return_value.cu.diff.return_value = 'diff'
        device = junos.ConnectedDevice(self.fqdn)
        callback = mock.Mock()
        callback.side_effect = RuntimeError
        with pytest.raises(RuntimeError):
            device.commit('config', COMMIT_MESSAGE, callback)

        assert callback.called
        mocked_junos_device.return_value.cu.commit.assert_not_called()

    @pytest.mark.parametrize('insert, expected', (
        ('', 'Commit error: Error Message'),
        ('<ok/>', 'Commit error: CommitError'),
    ))
    def test_commit_commit_error(self, mocked_junos_device, insert, expected):
        """On CommitError it should raise HomerError with the error message from the CommitError."""
        mocked_junos_device.return_value.cu = mock.MagicMock(spec_set=junos.Config)
        mocked_junos_device.return_value.cu.diff.return_value = 'diff'
        mocked_junos_device.return_value.cu.commit.side_effect = CommitError(
            etree.XML(ERROR_RESPONSE.format(insert=insert)))
        device = junos.ConnectedDevice(self.fqdn)
        callback = mock.Mock()

        with pytest.raises(HomerError, match=expected):
            device.commit('config', COMMIT_MESSAGE, callback)

        callback.assert_called_once_with('device1.example.com', 'diff')
        mocked_junos_device.return_value.cu.commit.assert_called_once_with(confirm=2, comment=COMMIT_MESSAGE)

    def test_commit_generic_error(self, mocked_junos_device):
        """On any other exception it should raise HomerError."""
        mocked_junos_device.return_value.cu = mock.MagicMock(spec_set=junos.Config)
        mocked_junos_device.return_value.cu.diff.return_value = 'diff'
        mocked_junos_device.return_value.cu.commit.side_effect = ValueError
        device = junos.ConnectedDevice(self.fqdn)
        callback = mock.Mock()

        with pytest.raises(ValueError):
            device.commit('config', COMMIT_MESSAGE, callback)

        callback.assert_called_once_with('device1.example.com', 'diff')
        mocked_junos_device.return_value.cu.commit.assert_called_once_with(confirm=2, comment=COMMIT_MESSAGE)

    def test_commit_check_ok(self, mocked_junos_device):
        """It should commit check the new config and clean rollback the staged one."""
        mocked_junos_device.return_value.cu = mock.MagicMock(spec_set=junos.Config)
        mocked_junos_device.return_value.cu.diff.return_value = 'diff'
        device = junos.ConnectedDevice(self.fqdn)

        success, diff = device.commit_check('config')

        assert success
        assert diff == 'diff'
        mocked_junos_device.return_value.cu.commit_check.assert_called_once_with()
        mocked_junos_device.return_value.cu.rollback.assert_called_once_with()

    def test_commit_check_commit_error(self, mocked_junos_device, caplog):
        """On CommitError it should log the error message from the CommitError."""
        mocked_junos_device.return_value.cu = mock.MagicMock(spec_set=junos.Config)
        mocked_junos_device.return_value.cu.diff.return_value = 'diff'
        mocked_junos_device.return_value.cu.commit_check.side_effect = CommitError(
            etree.XML(ERROR_RESPONSE.format(insert='')))
        device = junos.ConnectedDevice(self.fqdn)

        success, diff = device.commit_check('config')

        assert not success
        assert diff == 'diff'
        assert 'Commit check error on {fqdn}: Error Message'.format(fqdn=self.fqdn) in caplog.text
        mocked_junos_device.return_value.cu.commit_check.assert_called_once_with()
        mocked_junos_device.return_value.cu.rollback.assert_called_once_with()

    def test_commit_check_generic_error(self, mocked_junos_device, caplog):
        """On any other exception it should log the stacktrace from the exception."""
        mocked_junos_device.return_value.cu = mock.MagicMock(spec_set=junos.Config)
        mocked_junos_device.return_value.cu.diff.return_value = 'diff'
        mocked_junos_device.return_value.cu.commit_check.side_effect = ValueError('Error Message')
        device = junos.ConnectedDevice(self.fqdn)

        success, diff = device.commit_check('config')

        assert not success
        assert diff == 'diff'
        assert 'Failed to commit check on {fqdn}: Error Message'.format(fqdn=self.fqdn) in caplog.text
        mocked_junos_device.return_value.cu.commit_check.assert_called_once_with()
        mocked_junos_device.return_value.cu.rollback.assert_called_once_with()

    def test_commit_check_rollback_error(self, mocked_junos_device, caplog):
        """It should log any rollback error."""
        mocked_junos_device.return_value.cu = mock.MagicMock(spec_set=junos.Config)
        mocked_junos_device.return_value.cu.diff.return_value = 'diff'
        mocked_junos_device.return_value.cu.rollback.side_effect = ValueError(50)
        device = junos.ConnectedDevice(self.fqdn)

        success, diff = device.commit_check('config')

        assert success
        assert diff == 'diff'
        assert 'Invalid rollback ID on {fqdn}: 50'.format(fqdn=self.fqdn) in caplog.text
        mocked_junos_device.return_value.cu.commit_check.assert_called_once_with()
        mocked_junos_device.return_value.cu.rollback.assert_called_once_with()

    def test_close_ok(self, mocked_junos_device):
        """It should unlock and close the connection."""
        mocked_junos_device.return_value.cu = mock.MagicMock(spec_set=junos.Config)
        device = junos.ConnectedDevice(self.fqdn)

        device.close()

        mocked_junos_device.return_value.cu.unlock.assert_called_once_with()
        mocked_junos_device.return_value.close.assert_called_once_with()

    def test_close_unlock_error(self, mocked_junos_device):
        """On UnlockError it should continue to close the connection."""
        mocked_junos_device.return_value.cu = mock.MagicMock(spec_set=junos.Config)
        mocked_junos_device.return_value.cu.unlock.side_effect = UnlockError(
            etree.XML(ERROR_RESPONSE.format(insert='')))
        device = junos.ConnectedDevice(self.fqdn)

        device.close()

        mocked_junos_device.return_value.cu.unlock.assert_called_once_with()
        mocked_junos_device.return_value.close.assert_called_once_with()

    def test_close_timeout(self, mocked_junos_device):
        """On TimeoutExpiredError it should not fail."""
        mocked_junos_device.return_value.cu = mock.MagicMock(spec_set=junos.Config)
        mocked_junos_device.return_value.close.side_effect = TimeoutExpiredError
        device = junos.ConnectedDevice(self.fqdn)

        device.close()

        mocked_junos_device.return_value.cu.unlock.assert_called_once_with()
        mocked_junos_device.return_value.close.assert_called_once_with()



"""Homer's CLI."""
import argparse
import logging
import sys

from typing import Optional

from homer import Homer
from homer.config import load_yaml_config


def argument_parser() -> argparse.ArgumentParser:
    """Get the CLI argument parser.

    Returns:
        argparse.ArgumentParser: the argument parser instance.

    """
    parser = argparse.ArgumentParser(description="Configuration manager for network devices")
    parser.set_defaults(loglevel=logging.INFO)
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-v', '--verbose', action='store_const', const=logging.DEBUG, dest='loglevel',
                       help='Verbose (debug) logging')
    group.add_argument('-q', '--quiet', action='store_const', const=logging.WARN, dest='loglevel',
                       help='Silent mode, only log warnings',)
    parser.add_argument('-c', '--config', default='/etc/homer/config.yaml', help='Main configuration file to load.')
    parser.add_argument('query', help='Select which devices to target')

    subparsers = parser.add_subparsers(help='Action to perform: generate, diff, commit', dest='action')
    subparsers.required = True

    subparsers.add_parser('generate', help='Generate the configurations locally.')

    diff = subparsers.add_parser('diff', help=('Perform a commit check and show the differences between the generated '
                                               'configuration and the live one.'))
    diff.add_argument('-o', '--omit-diff', action='store_true',
                      help='Omit the actual diff to prevent the leak of private data')

    commit = subparsers.add_parser('commit', help='Actually commit the generated configuration to the devices.')
    commit.add_argument('message', help='A mandatory commit message. The running username will be automatically added.')

    return parser


def main(argv: Optional[list] = None) -> int:
    """Run the Homer CLI.

    Arguments:
        argv (list): the command line arguments to parse.

    Returns:
        int: ``0`` on success, ``1`` on failure.

    """
    args = argument_parser().parse_args(argv)
    logging.basicConfig(level=args.loglevel)
    if args.loglevel != logging.DEBUG:  # Suppress noisy loggers
        logging.getLogger('ncclient').setLevel(logging.WARNING)

    kwargs = {}
    if args.action == 'commit':
        kwargs['message'] = args.message
    elif args.action == 'diff':
        kwargs['omit_diff'] = args.omit_diff

    config = load_yaml_config(args.config)
    runner = Homer(config)
    return getattr(runner, args.action)(args.query, **kwargs)


if __name__ == '__main__':
    sys.exit(main(argv=sys.argv[1:]))



"""Custom plugin module."""
from homer.netbox import BaseNetboxDeviceData


class NetboxDeviceDataPlugin(BaseNetboxDeviceData):  # pylint: disable=too-many-ancestors
    """Custom class to mangle device-specific Netbox data."""

    def _get_netbox_device_plugin(self) -> str:
        """Example custom method that returns a string.

        Returns:
            str: True for testing purposes.

        """
        if self._device.metadata:
            return 'netbox_device_plugin'

        return ''



# -*- coding: utf-8 -*-
"""Sphynx configuration."""
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

from pkg_resources import get_distribution

import sphinx_rtd_theme

from sphinx import __version__ as sphinx_version

# Adjust path
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), os.pardir)))


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.viewcode',
    'sphinx.ext.githubpages',
    'sphinxarg.ext',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'Homer'
title = '{project} Documentation'.format(project=project)
copyright = ('2019, Riccardo Coccioli <rcoccioli@wikimedia.org>, Arzhel Younsi <ayounsi@wikimedia.org>, '
             'Faidon Liambotis <faidon@wikimedia.org>, Wikimedia Foundation, Inc.')
author = 'Riccardo Coccioli'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The full version, including alpha/beta/rc tags.
release = get_distribution('homer').version
# The short X.Y version.
version = release


# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True


# -- Options for HTML output ----------------------------------------------

html_theme = 'sphinx_rtd_theme'
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
sphinx_version_parts = [int(i) for i in sphinx_version.split('.')]
if sphinx_version_parts[0] == 1 and sphinx_version_parts[1] < 6:
    html_use_smartypants = False

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'Homerdoc'


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
# man_pages = [
# TODO
# ]


# -- Options for intersphinx ---------------------------------------

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
}

# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = False
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = False
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True
napoleon_use_keyword = True

# Autodoc settings
autodoc_default_flags = ['members', 'show-inheritance', 'inherited-members']
autoclass_content = 'both'


# -- Helper functions -----------------------------------------------------

def skip_external_inherited(app, what, name, obj, skip, options):
    """Skip methods inherited from external libraries."""
    if not (what == 'class' and hasattr(obj, 'im_class') and obj.__name__ not in obj.im_class.__dict__):
        return

    classes = [obj.im_class]
    while classes:
        c = classes.pop()
        if obj.__name__ in c.__dict__:  # Found the class that defined the method
            return 'homer' not in c.__module__  # Skip if from external libraries
        else:
            classes = list(c.__bases__) + classes  # Continue in the inheritance tree


def skip_exceptions_init(app, what, name, obj, skip, options):
    """Skip __init__ and with_traceback methods for Exception classes."""
    if what == 'exception' and name in ('__init__', 'with_traceback'):
        return True


def setup(app):
    """Register the filter_namedtuple_docstrings function."""
    app.connect('autodoc-skip-member', skip_exceptions_init)
    app.connect('autodoc-skip-member', skip_external_inherited)
    app.add_stylesheet('theme_overrides.css')  # override wide tables in RTD theme
